
require "util"

-- test harness relies on *global variable* cc
-- assumed also to be embedded in fn_to_test.

-- You will find make_collector defined in 
-- the *new* version of "init.lua" 
-- in this directory.   

cc = util.make_collector() 
   
-- 

function count_test_harness (fn_to_test, test_arrays) 
   local result = {} 
   for i = 1, #test_arrays do 
      cc.reset() 
      fn_to_test(test_arrays[i]) 
      result[i] = cc.reveal() 
      cc.reset() 
      io.write(".") ; io.flush() 
   end 
   return result 
end 

function make_unimodal_test_arrays(number, width, upper_bound)
   local result = {} 
   for i = 1, number do 
      result[i] = make_unimodal (width, upper_bound)
   end 
   return result 
end 

-- ----------------
-- unimodal arrays 
-- ----------------

-- First, a routine to generate unimodal 
-- sample arrays.  

function make_unimodal (width, upper_bound) 
   local max_val = math.random(upper_bound) -- random maximum 
   local pos_max = math.random(width)       -- random position to put it 
   local result = {} 

   if pos_max<=1                -- insert the first entry 
   then
      result[1]=max_val    
   else 
      result[1] = 0
   end 

   for i = 2, width do          -- insert remaining entries 
      if i < pos_max 
      then                      -- increase last entry by a 
         result[i] = result[i-1] + -- reasonable fraction of whats left to max_val
            ((max_val - result[i-1])/(pos_max-i+1))
      else if i == pos_max 
         then                   -- time to insert max_val 
            result[i] = max_val
         else                   -- decrease proportionally 
            result[i] = max_val - 
               ((i-pos_max)*max_val / (width-pos_max))
         end 
      end 
   end 
   return result                -- return the array we built 
end 

-- next, a routine to check if array a 
-- is unimodal.  When it stops increasing, 
-- it had better never increase again.  

function is_unimodal(a) 
   local inc_part = true 
   for i = 2, #a do 
      if a[i] == a[i-1] then return false -- repeated value is not increase! 
      else if a[i-1] > a[i]
         then 
            if inc_part 
            then inc_part=false -- not in increasing part any more!
            end
         else                   -- a[i] is greater! 
            if not(inc_part) 
            then return false   -- decrease failed
            end
         end 
      end 
   end 
   return true                  -- got to the end with nothing wrong 
end 
               
-- The function to find the maximum 
-- position.  Only has to work if 
-- the given array *a* is really unimodal.  

-- Please fill this in using a divide-and-conquer  
-- idea.  See instructions in proj2.pdf 

function uni_max_rec (a,left,right) 

-- Your code goes here!  

end 

-- a convenient top-level function to 
-- start off uni_max_rec with the natural 
-- left and right values.  

function uni_max(a) 
   return uni_max_rec(a,1,#a)
end 

-- Here is a function that defines 
-- the idea of a local maximum.  

-- assume 1 <= lower <= pos <= upper <= #a.  
-- 
-- Then is_local_max(a,pos,lower,upper) returns true 
-- just when a[pos] is strictly greater than a[i] 
-- for all i adjacent to pos and within
-- lower <= i <= upper.  

function is_local_max(a,pos,lower,upper) 
   if lower == upper then return true 
   else if pos == lower then return a[pos+1]<a[pos] 
      else if pos == upper then return a[pos-1]<a[pos] 
         else return (a[pos-1]<a[pos] and a[pos+1]<a[pos])
         end 
      end
   end 
end 

  

-- maximum subarray 


-- Here is the brute-force procedure.  
-- It should be simple to write, and it 
-- does not have to be efficient.  Use it 
-- to compare timings, and especially to make 
-- sure that your recursive, divide-and-conquer 
-- algorithm is giving the same results.  

function find_maximum_subarray_brute_force(a, left, right)

   -- place your code here!

end 

-- Procedure to find a maximum array that 
-- lies within left and right, and actually 
-- crosses the midpoint mid.  

function find_max_crossing(a,left,mid,right) 
   
   -- Place your Theta(right-left) code 
   -- to compute the start, end, and sum of the 
   -- maximum array that crosses the midpoint.

   -- return your information in this form, 
   -- to make available all three parts: 

   return max_start, max_end, (left_val + right_val)
end 
   

-- Here is the recursive procedure to fill in, 
-- using a divide-and-conquer strategy.  Its 
-- runtime should end up as Theta(n log n).  


function find_maximum_subarray_rec(a, left, right) 

   -- place your code here!
   -- You can use this format to retrieve 
   -- multiple values from a recursive call: 

   local l_start, l_end, l_max_internal = 
      find_maximum_subarray_rec(a,left,mid)
end 

-- Here is a convenient top-level call 
-- to start the recursive procedure on the 
-- array as a whole.

function find_maximum_subarray(a)
   return find_maximum_subarray_rec(a, 1, #a)
end 

-- here is a function you can use to generate 
-- a random array of real numbers.  The resulting 
-- array will be of length "length".  Its entries 
-- will be bounded in absolute value by max_abs.

function make_random_real_array(length, max_abs) 
   local result = {} 
   local m = 2*max_abs 

   for i = 1, length do 
      result[i] = (math.random() - .5) * m
   end
   return result 
end 

-- you can use this procedure to generate random arrays and to test
-- find_maximum_subarray and find_maximum_subarray_brute_force on
-- them.  

-- When it finds an array where they give different values, it raises
-- an error, and it stores the array in the global variable "array".  

function test_max_subarray(length, max_abs, num) 
   if 0 < num then 
      local a = make_random_real_array(length, max_abs)
      local s,e,m = find_maximum_subarray(a,1,#a) 
      local s1,e1,m1 = find_maximum_subarray_brute_force(a,1,#a)
      if s == s1 and e == e1 then write_dot() else 
         array = a 
         print(" found failure, stored in array")
      end
      test_max_subarray(length, max_abs, num-1)
   else 
      print("done.")
   end 
end 
